#!/bin/bash


# To be consistent, all functions should return
# 0=TRUE, 1=FALSE
#
#

# No Operation Flag (noop)
#
# Enable or disable script from modifying filesystem.
# Set this to 1 for testing the script.
# set to 0 for real operations.
noop=0



############
function fatal()
{
    c0=`caller 0`
    c=`caller 1`
    #shift 1
    echo -e "Fatal error $c: $*" >&2
    echo "Was Called by $c0";
    exit 1
}

# Interpet exit status as boolean and invert the meaning

function notop()
{
    if [ $? -lt 0 ];
	then return 1;
	else return 0;
    fi
}

#
# File Related Library Functions
#

function _FileIs()
{
    if [ $# != 2 ]; then fatal not enough arguments.; fi

    filename=$2

    case $1 in 
	regular) cond="-f" ;;
	link) cond="-h" ;;
	dir)  cond="-d" ;;
    esac

    if [ $cond $filename ]; then return 0; fi
    return 1
}

function ambiguous()
{
    fatal $1 this ambigous command is disabled.\n \
valid functions: \
isdir \
islink \
isfile \
cp_rf \
mv_f \
mkdir_p \
\n\
rm_rf \
rm_link \
rm_linkOrDir \
ln_sf \
chmod \ 
verify_access \

}

##
## isdir()
## Returns 0=TRUE 1=FALSE
##
function isdir()
{
    if [ $# != 1 ]; then fatal not enough arguments.; fi
    _FileIs dir $1
    return $?
}
alias verify_link=isdir

## islink()
## Returns 0=TRUE 1=FALSE
##
function islink()
{
    if [ $# != 1 ]; then fatal not enough arguments.; fi
    _FileIs link $1
    return $?
}
alias verify_link=islink

## isfile()
## Test for regular file (not dirs, links, etc)
## Returns 0=TRUE 1=FALSE
##
function isfile()
{
    if [ $# != 1 ]; then fatal not enough arguments.; fi
    _FileIs regular $1
    return $?
}
alias verify_file=isfile

## cp_rf()
##
## Returns 0=OK, 1=ERROR
##

alias cp="ambiguous cp"
function cp_rf()
{
    if [ $# -lt 2 ]; then fatal not enough arguments.;
    fi
    cmd="cp -rf $@"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

## mv_f()
##
## Returns 0=OK, 1=ERROR
##

alias mv="ambiguous mv"
function mv_f()
{
    if [ $# -lt 2 ]; then fatal not enough arguments.;
    fi
    cmd="mv -f $@"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

## mkdir_p()
##
## Returns 0=OK, 1=ERROR
##

alias mkdir="ambiguous mkdir"
function mkdir_p()
{
    if [ $# -lt 1 ]; then fatal not enough arguments.;
    fi
    cmd="mkdir -p $@"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

## rm_rf()
##
## Returns 0=OK, 1=ERROR
## Appears to always return 0

alias rm="ambigous rm"
function rm_rf()
{
    if [ $# -lt 1 ]; then fatal not enough arguments.;
    fi

# DO ALLOW USE ON LINKS!!
#    # This only catches cases of "rm_rf LINK", not "rm_rf FILE1 FILE2 LINK"
#    if [ -h $1 ];
#	then
#	 fatal do not use this on links lik $1. Use rm_link instead.
#    fi

    cmd="rm -rf $@"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

## Do not use this on actual directories.

function rm_link()
{
    // DO NOT use recursive remove here!!

    if [ $# -lt 1 ]; then fatal not enough arguments.;
    fi
    cmd="rm -f $@"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

# rm_linkOrDir only takes a single arg, no wildcards
# Returns 0=OK 1=FAILED
#
function rm_linkOrDir()
{
    target=$1

    islink $target \
	|| rm_link $target

    isdir $target \
	|| rm_rf $target

    isdir $target \
	|| islink $target

    ret=notop
    return $?
}

## ln_sf()
##
## Returns 0=OK, 1=ERROR

alias ln="ambiguous ln"
function ln_sf()
{
    if [ $# -ne 2 ]; then fatal not enough arguments.;
    fi

    islink $2
    lyes=$?
    isfile $2
    fyes=$?


    if [[ $lyes -eq 0 || $fyes -eq 0 ]];
	then
	echo "Link Target ${2} aleady exists. Removing it..."
	rm_rf $2
    fi

    cmd="ln -sf $1 $2"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

# More of a testing func than anything
function xxls()
{
    if [ $# != 1 ]; then fatal not enough arguments.;
    fi
    cmd="ls $1"
    echo "$cmd"; if [ $noop -gt 0 ]; then return 0; fi
    echo "in real"
    command $cmd
    return $?
}

## chmod()
##
## Returns 0=OK, 1=ERROR

function chmod()
{
    if [ $# -lt 2 ]; then fatal not enough arguments.;
    fi
    cmd="chmod $@"
    echo "$cmd";  if [ $noop -gt 0 ]; then return 0; fi
    command $cmd
    return $?
}

#
# verify_access
#
# Example call:  verify_access 1 r-- root rw- r-- FOO
# This will verify that file FOO has
# group=1, r--
# owner=root, rw-
# others have r-- 
#
# returns 0=OK 1=BAD

function verify_access()
{
    if [ $# != 6 ]; then fatal not enough arguments.; fi

    expectGroup=$1
    expectGroupBits=$2
    expectOwner=$3
    expectOwnerBits=$4
    expectOtherBits=$5

    file=$6
    echo "verify_access($file)"
    lsline=`ls -l $file`
    bits=`echo $lsline | cut -c 2-10`

    OwnerBits=`echo $bits|cut -c 1-3`
    OtherBits=`echo $bits|cut -c 4-6`
    GroupBits=`echo $bits|cut -c 7-9`

    Group=`echo $lsline | awk '{ print $2; }'`
    Owner=`echo $lsline | awk '{ print $3; }'`
    Creator=`echo $lsline | awk '{ print $4; }'`
    
    echo "found  : $bits "$Group " " $Owner
    echo "expect : "$expectOwnerBits$expectGroupBits$expectOtherBits" "$expectGroup " " $expectOwner


    err=0

    if [ $expectOwnerBits != $OwnerBits ]; then (( err++ )); fi
    if [ $expectOtherBits != $OtherBits ]; then (( err++ )); fi
    if [ $expectGroupBits != $GroupBits ]; then (( err++ )); fi

    if [ $expectOwner != $Owner ]; then (( err++ )); fi
    ## if [ $expectCreator != $Creator ]; then (( err++ )); fi
    if [ $expectGroup != $Group ]; then (( err++ )); fi

    if [ $err -eq 0 ];
	then
	  echo "Access OK";
	else
	  echo "Access BAD";
    fi
    return $err
}


#### Link Tracking Stuff - Public Funcs
#
# track_ln_sf  -- Create a link.
# lib_check_links -- check links (see important notes below)
# lib_repair_links -- repair and verify repair (see important notes below)
#
#

TRACKMODE=0
ln_repair_mode=0
ln_repair_err=0
check_ln_pairs_err=0
glob_check_source=""
glob_check_source_used=0

typeBadDir=1
typeBadFile=2
typeBadLink=3
typeMispointLink=4

function track_ln_sf()
{
#    funcname=$1
#    shift 1

    if [ $TRACKMODE -eq 1 ];
    then
	# print input args for calling func
	# do not modify this line!
	echo "$@"
	return 0
    fi;

    # Call the real ln_sf for non track mode
    ln_sf $1 $2

    return 0;
}


function _set_ln_repair_mode()
{
    ln_repair_mode=$1
}

# check_ln_pairs LIST
# Where list contains pairs DIR, TARGET_LINK
# Will check each pair and report on findings.
#
# Returns 0=Good 1=Bad

function check_ln_pairs()
{
    check_ln_pairs_err=0
    _check_ln_pairs $@
    return $check_ln_pairs_err
}

function status_to_str()
{
    m="exists"
    case $1 in 
	$typeBadDir) m="is not a directory";;
	$typeBadFile) m="does not exist";;
        $typeBadLink) m="does not exist";;
        $typeMispoitLink) m="points to wrong source";;
    esac

    echo $m

}
function _check_ln_pairs()
{

    while [ ${#*} -gt 1 ];
      do

      source=$1
      target=$2
      shift 2

      reallyPointsTo=""

      sourceSTATUS=0
      isdir $source
      if [ $? -ne 0 ];
	  then
	  sourceSTATUS=$typeBadDir
	  isfile $source
	  if [ $? -ne 0 ];
	      then
	      sourceSTATUS=$typeBadFile
	  fi
      fi

      islink $target
      if [ $? -ne 0 ];
	  then
	  targetSTATUS=$typeBadLink
      else
	  targetSTATUS=0
      fi
##
      status="OK  ";

      if [[ $sourceSTATUS -ne 0 || $targetSTATUS -ne 0 ]];
	  then
	  status="BAD "
	 let check_ln_pairs_err+=1
      fi


      reallyPointsTo=`readlink $target`
      msg=""

    # Check if link points to right location if link exists
      if [ $targetSTATUS -eq 0 ];
	then

	  # ignore trailing slash
	  reallyPointsTo=`echo $reallyPointsTo | sed "s/\/$//"`

	  if [ "$reallyPointsTo" != $source ];
	      then
	      msg="\n                  Mis-point: $reallyPointsTo != $source"
	      targetSTATUS=$typeMispointLink
	      status="BAD"
	      let check_ln_pairs_err+=1
	  fi;
      fi;

#echo -e "{$targetSTATUS $sourceSTATUS \
#r_mode=${ln_repair_mode}/PASS=${ln_pass}/pairs_err=${check_ln_pairs_err}/rep_err=${ln_repair_err}} $target->$source"

      mst=0
      m="         "
      ms=" "
      mt=" "
      if [ $sourceSTATUS -ne 0 ]; then m="ERROR:  "; ms="*"; mst=1; fi
      if [ $targetSTATUS -ne 0 ]; then m="ERROR:  "; mt="*"; mst=1; fi

      if [ $ln_repair_mode -eq 0 ];
	  then
	  # -------------- Informative Messaging


	  # Print a nice easy to read status message.
	  # there were complaints about the terse N->N format

	StatStrA=`status_to_str $sourceSTATUS`
	StatStrB=`status_to_str $targetSTATUS`

	if [[ $mst -eq 1 && $glob_check_source_used -eq 0 ]];
	    then
	    echo "$m ${glob_check_source} FAILED checks"
	    echo "$m"
	    glob_check_source_used=1
	fi
	echo "$m { ${ms}SOURCE: $source"
	echo "$m { ${mt}TARGET: $target"
	echo "$m |"
	echo "$m '--> Source directory ${StatStrA}, Target link ${StatStrB}";
	echo "$m "
#	  echo -e "    [$status $targetSTATUS->$sourceSTATUS]  $target -> $source$msg";


      else
	  # ATTEMPT TO REPAIR

	  if [ $ln_repair_mode -eq 1 ]; then
	      what="";
	  else
	      what="verification";
	  fi
	if [[ $mst -eq 1 && $glob_check_source_used -eq 0 ]];
	    then
	    echo "$m"
	    echo "$m Repair of ${glob_check_source} FAILED $what"
	    if [ $ln_repair_mode -eq 1 ]; then
		echo "$m"
	    fi
	    glob_check_source_used=1
	fi

	  if [ $ln_repair_mode -eq 1 ];
	      then

	      if [[ $mst -eq 1 && $glob_check_source_used -eq 0 ]];
		  then
		  echo "REPAIRING ${glob_check_source}..."
		  glob_check_source_used=1
	      fi

	      if [ $sourceSTATUS -ne 0 ];
		  then
		  echo "ERROR: Unable to repair missing dir: $source"
	      elif [ $targetSTATUS -eq $typeMispointLink ];
		  then

		  # link is pointing to the wrong directory.
		  # cannot delete link to an existing dir, so we need to
		  # force it here.

		  pathparts=`echo $reallyPointsTo|sed "s/\//\n/g"|wc -l`
		  root=`echo $reallyPointsTo | egrep "(root|ssh)"`

		  if [[ $pathparts -gt 2 && $root -eq 1 ]];
		      then
		      mv_f $reallyPointsTo $reallyPointsTo.tmp
		      track_ln_sf $source $target
		      mv_f $reallyPointsTo.tmp $reallyPointsTo
		  else
		      echo "ERORR: Unable to fix mispointed link: $target"
		      echo "ERROR: Requires manual deletion or fix of dir: $reallyPointsTo"
		  let ln_repair_err=ln_repair_err+=1
		  fi
	      elif [ $targetSTATUS -eq $typeBadLink ];
		  then
		      echo "Repairing link $target"
		      save=$TRACKMODE
		      TRACKMODE=0
		      track_ln_sf $source $target
		      TRACKMODE=$save

		  # check status
#		      let ln_repair_err=ln_repair_err+$saveerr
	      fi

	  else # CHECK REPAIR (ln_repairmode >1)
	      if [[ $sourceSTATUS -eq 0 && $targetSTATUS -eq 0 ]];
		  then
		  a=0 # Success. do nothing
	      else
		  let ln_repair_err=ln_repair_err+1;
		  a=0 # Failed. do nothing
	      fi
	  fi
      fi

      if [ ${#*} -eq 0 ];
		  then
	  break

      fi
    done

    if [ $ln_repair_mode -eq 1 ]; # ignore ln_repair_mode==2
	then
	return $ln_repair_err
    else
		return $check_ln_pairs_err
    fi
}



# Call this when you want links checked
# The script sourcing this Lib must contain function(s)
# named [i|u].*_link[s]
# These functions should ONLY contain calls to track_ln_sf
# repair_links() deals with link-directory relationships
# specified in track_ln_sf calls.
#
# Returns 0=success 1=Failed

function force_check_fail()
{
return 1
}
function lib_check_links()
{
    check_links_err=0
    err=0

    # could use 'declare -f'
    ln_funcs=`grep "function [iu].*_lin[ks]\(\)" $0 | tr -d ")" |tr -d "("`
    ln_funcs=`echo "$ln_funcs" | awk '{print $2;}'`


    savetrackmode=$trackmode
    TRACKMODE=1

    for item in $ln_funcs
    do
      if [ $ln_repair_mode -eq 0 ];
	  then
	  ignore=0
	 echo -e "\n$item :"
      fi
      a=`eval $item`


      glob_check_source=$item
      glob_check_source_used=0
      check_ln_pairs $a

      err=$?


      if [ $err -ne 0 ];
	  then
	  let check_links_err=check_links_err+1
     fi

    done
    TRACKMODE=$savetrackmode

    if [ $ln_repair_err -ne 0 ];
	then
	if [ $ln_repair_err -gt 1 ]; then s="s"; else s=""; fi
	echo "Repair error${s}.";
	return $ln_repair_err
    fi
    if [ $check_links_err -ne 0 ];
	then
	if [ $check_links_err -gt 1 ]; then s="s"; else s=""; fi
	echo "${check_links_err} group${s} of errors."
	return $check_links_err
    fi

    
    case $ln_repair_mode in 
	0) echo "No Check Errors";;
	1) echo "No Repair errors";;
        2) echo "Repair was Verified";;
    esac
    return 0
}

# Call this when you want links auto repaired
# The script sourcing this Lib must contain function(s)
# named [i|u].*_link[s]
# These functions should ONLY contain calls to track_ln_sf
# repair_links deals with link-directory relationships
# specified in track_ln_sf calls.
#
# Returns 0=success 1=Failed

function lib_repair_links()
{
    _set_ln_repair_mode 1
    ln_repair_err=0
    lib_check_links

    # apparently you can't easily manage recussive function calls in sh
    # so we have to re-enter the func from here
    # verify
   _set_ln_repair_mode 2

    lib_check_links
    err=$?
    return $err
}


#
# preserver_dir  CURRENT_LOCATION  NEW_LOCATION
# Returns: 0=Success 1=Failed

function preserve_dir()
{
    toPreserve=$1
    preserveLocation=$2

    if [ -d $toPreserve ]
        then

	# Important to remove target location if it exists
	# mv -f will not replace an existing directory

	rm_rf ${preserveLocation}

        mv_f $toPreserve ${preserveLocation}
        return $?;
    fi
    return 1;
}


